from . import compatibility, config, delegation, remote

VERSION: str = "0.4.1"
"""
Current version of QuickClone
"""

NAME: str = "QuickClone"
"""
Name of the project/app.
"""

DESCRIPTION: str = "Command line utility for quickly cloning remote SCM repositories as succintly \
as possible."
"""
A description of the app.
"""

